#include "Joystick.h"
#include "config.h"

LPDIRECTINPUT8 Joystick::g_DirectInput = NULL;
LPDIRECTINPUTDEVICE8 Joystick::m_Device = NULL;

Joystick::Joystick() {
    m_BumpEffect = NULL;

    pos[0] = pos[1] = 0;
}

Joystick::~Joystick() {
  printf("deleting Joystick (%x)\n", this);
}

bool Joystick::Init(HWND hwnd) {
    HRESULT hr;

    // Register with Direct Input
	hr = DirectInput8Create(GetModuleHandle(NULL), DIRECTINPUT_VERSION, IID_IDirectInput8, (void**) &g_DirectInput, NULL);
    if (FAILED(hr)) {
		printf("DirectInput8Create Failed.\n");
        return false;
	}

    // Find a Joystick
    hr = g_DirectInput->EnumDevices(DI8DEVCLASS_GAMECTRL, EnumJoysticksCallback, NULL, DIEDFL_ATTACHEDONLY | DIEDFL_FORCEFEEDBACK);
    if (FAILED(hr) || !m_Device) {
        printf("No Joystick Found.\n");
        return false;
    }

    // Tell it we want DIJOYSTATE-style data reports
	hr = m_Device->SetDataFormat(&c_dfDIJoystick);
    if (FAILED(hr)) {
        printf("Couldn't set data format.\n");
        return false;
    }

    // Set cooperative level
    hr = m_Device->SetCooperativeLevel(hwnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND);
	if (FAILED(hr)) {
		printf("Couldnt't set joystick cooperative level.\n");
		return false;
	}

    // Set up axis reporting
	hr = m_Device->EnumObjects(EnumAxesCallback, this, DIDFT_AXIS);
	if (FAILED(hr)) {
		printf("Can't enumerate joystick axes.\r\n");
	}

    // Since we will be playing force feedback effects, we should disable the
    // auto-centering spring.
    DIPROPDWORD prop;
    prop.diph.dwSize       = sizeof(DIPROPDWORD);
    prop.diph.dwHeaderSize = sizeof(DIPROPHEADER);
    prop.diph.dwObj        = 0;
    prop.diph.dwHow        = DIPH_DEVICE;
    prop.dwData            = FALSE;

    hr = m_Device->SetProperty(DIPROP_AUTOCENTER, &prop.diph);
    if (FAILED(hr)) {
        printf("Couldnt turn off auto-center.\n");
        return false;
    }

    m_Device->Acquire();
    if( !InitBumpEffect()) {
        printf("Couldn't create effects.\n");
        return false;
    }
    m_Device->Acquire();


    return true;
}

void Joystick::Update() {
    HRESULT hr;

	hr = m_Device->Poll();
	if (FAILED(hr)) {
		hr = m_Device->Acquire();
		while (hr == DIERR_INPUTLOST)
			hr = m_Device->Acquire();

		if (FAILED(hr)) {
			printf("Can't poll joystick device.\n");
			return;
		}
	}

	hr = m_Device->GetDeviceState(sizeof(DIJOYSTATE), &m_State);
	if (FAILED(hr)) {
		printf("Can't read joystick state.\n");
	}

    pos[0] = (double)m_State.lX / (double)AXIS_MAX;
    pos[1] = (double)m_State.lY / (double)AXIS_MAX;
}

void Joystick::GetPosition(double* position) {
    position[0] = pos[0];
    position[1] = pos[1];
}

void Joystick::PlayBump(int wall) {
    //m_BumpUpEffect->Start(1, 0);
    DoBumpEffect(wall);
}

BOOL Joystick::InitBumpEffect() {
	DWORD m_Axes[2];
	LONG m_Dirs[2];
    DICONSTANTFORCE m_Force;
    LPDIENVELOPE m_Envelope = NULL;
    DIEFFECT m_Effect;

	m_Axes[0] = DIJOFS_X;
	m_Axes[1] = DIJOFS_Y;
	m_Dirs[0] = 0;
	m_Dirs[1] = 1;

    m_Force.lMagnitude = config::forceFeedbackMagnitude;

	ZeroMemory(&m_Effect, sizeof(m_Effect));
	m_Effect.dwSize = sizeof(m_Effect);
	m_Effect.dwFlags = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS;
	m_Effect.dwDuration = DI_SECONDS/90;
	m_Effect.dwSamplePeriod = 0;
	m_Effect.dwStartDelay = 0;
    m_Effect.dwGain = config::forceFeedbackGain;
	m_Effect.dwTriggerButton = DIEB_NOTRIGGER;
	m_Effect.dwTriggerRepeatInterval = 0;
	m_Effect.cAxes = 2;
	m_Effect.rgdwAxes = m_Axes;
	m_Effect.rglDirection = m_Dirs;
	m_Effect.lpEnvelope = NULL;
	m_Effect.cbTypeSpecificParams = sizeof(m_Force);
	m_Effect.lpvTypeSpecificParams = &m_Force;

    HRESULT hr = m_Device->CreateEffect(GUID_ConstantForce, &m_Effect, &m_BumpEffect, NULL);
    if (FAILED(hr)) {
        printf("CreateEffect failed.\n");
        return false;
    }

    return true;
}

//-----------------------------------------------------------------------------
// Name: SetDeviceForcesXY()
// Desc: Apply the X and Y forces to the effect we prepared.
//-----------------------------------------------------------------------------
HRESULT Joystick::DoBumpEffect(int wall) {
    // Modifying an effect is basically the same as creating a new one, except
    // you need only specify the parameters you are modifying
    LONG m_Dirs[2] = {0, 0};

    DICONSTANTFORCE m_Force;

    m_Dirs[0] = m_Dirs[1] = 0;
    if (wall == WALL_LEFT) {
        m_Dirs[0] = -1000;
    }
    else if (wall == WALL_RIGHT) {
        m_Dirs[0] = 1000;
    }
    else if (wall == WALL_UP) {
        m_Dirs[1] = -1000;
    }
    else {
        m_Dirs[1] = 1000;
    }

    m_Force.lMagnitude = config::forceFeedbackMagnitude;

    DIEFFECT m_Effect;
    ZeroMemory( &m_Effect, sizeof(m_Effect) );
    m_Effect.dwSize                = sizeof(DIEFFECT);
    m_Effect.dwFlags               = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS;
    m_Effect.cAxes                 = 2;
    m_Effect.rglDirection          = m_Dirs;
    m_Effect.lpEnvelope            = 0;
    m_Effect.cbTypeSpecificParams  = sizeof(DICONSTANTFORCE);
    m_Effect.lpvTypeSpecificParams = &m_Force;
    m_Effect.dwStartDelay          = 0;

    // Now set the new parameters and start the effect immediately.
    return m_BumpEffect->SetParameters( &m_Effect, DIEP_DIRECTION |
                                           DIEP_TYPESPECIFICPARAMS |
                                           DIEP_START );
}

// Joystick enumerator. Add the joystick details to the list.
BOOL Joystick::EnumJoysticksCallback(const DIDEVICEINSTANCE *pInst, void *pContext) {
    LPDIRECTINPUTDEVICE8 pDevice;
    HRESULT              hr;

    // Obtain an interface to the enumerated force feedback device.
    hr = g_DirectInput->CreateDevice(pInst->guidInstance, &pDevice, NULL);

    // If it failed, then we can't use this device for some
    // bizarre reason.  (Maybe the user unplugged it while we
    // were in the middle of enumerating it.)  So continue enumerating
    if(FAILED(hr)) {
        return DIENUM_CONTINUE;
    }

    // We successfully created an IDirectInputDevice8.  So stop looking 
    // for another one.
    m_Device = pDevice;

    return DIENUM_STOP;
}

// Axis Enumerator.  Change axis range
BOOL CALLBACK Joystick::EnumAxesCallback(LPCDIDEVICEOBJECTINSTANCE lpddoi, void *pvRef) {
	// Convert axis range to [AXIS_MIN, AXIS_MAX]
	DIPROPRANGE diprg;

	diprg.diph.dwSize = sizeof(DIPROPRANGE);
	diprg.diph.dwHeaderSize = sizeof(DIPROPHEADER);
	diprg.diph.dwHow = DIPH_BYID;
	diprg.diph.dwObj = lpddoi->dwType;
	diprg.lMin = AXIS_MIN;
	diprg.lMax = AXIS_MAX;
	if (FAILED(m_Device->SetProperty(DIPROP_RANGE, &diprg.diph))) {
		printf("Unable to set axis property.\n");
		return DIENUM_STOP;
	}

	return DIENUM_CONTINUE;
}