#include "MazeGraphics.h"
#include "config.h"

MazeGraphics::MazeGraphics() {
}

MazeGraphics::MazeGraphics(Maze* m, Player* p) {
    maze = m;
    player = p;

    Init();
}

MazeGraphics::~MazeGraphics() {
  printf("deleting MazeGraphics (%x)\n", this);
}

void MazeGraphics::Init() {
    // Create the maze graphics
    CreateMazeList();
    CreateMazeStartList();
    CreateMazeExitList();
    CreateMazeFloorList();
    CreatePlayerList();
}

void MazeGraphics::Update(double elapsed) {
     DrawGraphics();
}

void MazeGraphics::CreateMazeList() {
    glDeleteLists(mazeList, 1);
    mazeList = glGenLists(1);

    glNewList(mazeList, GL_COMPILE);
    glEnable(GL_COLOR_MATERIAL);
    glColor3f(1, 1, 1);
    // Draw maze blocks
    for (int i = 0; i < maze->getWidth(); i++) {
        for (int j = 0; j < maze->getHeight(); j++) {
            // Regular wall
            if (maze->getCell(i, j)->getStatus() == Cell::GENERIC_WALL) { 
                glColor3f(1, 1, 1);
                glPushMatrix();
                glTranslatef(i + 0.5, 0, j + 0.5);
                glutSolidCube(1);
                glPopMatrix();
            }
        }
    }
    glEndList();
}

void MazeGraphics::CreateMazeStartList() {
    // Get start position
    double start[2];
    maze->getStart(start);

	glDeleteLists(mazeStartList, 1);
    mazeStartList = glGenLists(1);

    glNewList(mazeStartList, GL_COMPILE);
    glEnable(GL_COLOR_MATERIAL);
    glColor4f(0, 1, 0, 0.5);
    glPushMatrix();
    glTranslatef(start[0], 0, start[1]);
    glutSolidCube(1);
    glPopMatrix();
    glEndList();
}

void MazeGraphics::CreateMazeExitList() {
	glDeleteLists(mazeExitList, 1);
    mazeExitList = glGenLists(1);

    glNewList(mazeExitList, GL_COMPILE);
    glEnable(GL_COLOR_MATERIAL);
    glColor4f(1, 0, 0, 0.5);
    // Draw exit blocks
    for (int i = 0; i < maze->getWidth(); i++) {
        for (int j = 0; j < maze->getHeight(); j++) {
            if (maze->getCell(i, j)->getStatus() == Cell::EXIT) { 
                glPushMatrix();
                glTranslatef(i + 0.5, 0, j + 0.5);
                glutSolidCube(1);
                glPopMatrix();
            }
        }
    }
    glEndList();
}

void MazeGraphics::CreateMazeFloorList() {
    glDeleteLists(mazeFloorList, 1);
    mazeFloorList = glGenLists(1);

    glNewList(mazeFloorList, GL_COMPILE);
    glEnable(GL_COLOR_MATERIAL);

    glBegin(GL_QUADS);
    for (int i = 0; i < maze->getWidth(); i++) {
      for (int j = 0; j < maze->getHeight(); j++) {
        Zone * zone = maze->getZone(i, j);
        if (zone != NULL)
          glColor3ub(zone->color & 0xFF, (zone->color >> 8) & 0xFF, zone->color >> 16);
        else
          glColor3f(0.9, 0.9, 1.0);
        glVertex3f(i, -0.5, j);
        glVertex3f(i, -0.5, j+1.0);
        glVertex3f(i+1.0, -0.5, j+1.0);
        glVertex3f(i+1.0, -0.5, j);
      }
    }
    glEnd();

/*    // Draw floor
        glVertex3f(0, -0.5, 0);
        glVertex3f(0, -0.5, maze->getHeight());
        glVertex3f(maze->getWidth(), -0.5, maze->getHeight());
        glVertex3f(maze->getWidth(), -0.5, 0);
    glEnd();*/
    glEndList();
}

void MazeGraphics::CreatePlayerList() {
	glDeleteLists(playerList, 1);
    playerList = glGenLists(1);
    glNewList(playerList, GL_COMPILE);

    glEnable(GL_COLOR_MATERIAL);

    // Sphere
    glColor3f(0, 0, 1);
    glutSolidSphere(0.1, 16, 16);

    // Cone
    double coneHeight = playerHeight + 0.5;
    glPushMatrix();
    glTranslatef(0, -coneHeight, 0);
    glRotatef(-90, 1, 0, 0);
    glutSolidCone(0.2, coneHeight, 16, 16);
    glPopMatrix();

    // Halo
    double haloHeight = playerHeight + 0.25;
    glColor3f(1, 1, 0);
    glPushMatrix();
    glTranslatef(0, haloHeight, 0);
    glRotatef(90, 1, 0, 0);
    glutSolidTorus(0.02, 0.1, 2, 16);
    glPopMatrix();

    glEndList();
}

void MazeGraphics::DrawGraphics() {
    double cameraPos[3] = {(double)maze->getWidth() / 2.0, (double)maze->getWidth() * 0.9, (double)maze->getWidth() * 1.2};

    // Place the camera
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // Create the lighting
    static const GLfloat light0_diffuse[] = {1, 1, 1, 1};   
    static const GLfloat light0_position[] = {0, 0, 0, 1}; 

    glDisable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glShadeModel(GL_SMOOTH);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
    glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
    glEnable(GL_LIGHT0); 

    gluLookAt(cameraPos[0], cameraPos[1], cameraPos[2],
             (double)maze->getWidth() / 2.0, 0, (double)maze->getHeight() / 2.0,
             0, 1, 0);

    // Don't draw if screen blanked
    if (!config::screenBlanked) {
        glPushMatrix();
        double playerPos[3];
        player->GetPosition(playerPos);
        glTranslatef(playerPos[0], playerPos[1], playerPos[2]);
        glCallList(playerList);
        glPopMatrix();

        glCallList(mazeStartList);
        glCallList(mazeExitList);
        glCallList(mazeList);
        glCallList(mazeFloorList);
    }
}