#ifdef PHANTOM_ENABLED

//////////////////////////////////////////////////////////////////////

#include "MazeGraphicsHaptics.h"
#include "config.h"

//////////////////////////////////////////////////////////////////////

MazeGraphicsHaptics::MazeGraphicsHaptics(Maze* m, Player* p) {
    maze = m;
    player = p;

    atStart = false;

    friction = hlGenEffects(1);

    Init();

    // Generate id's for the haptic geometry
    mazeListId = hlGenShapes(1);
    constraintListId = hlGenShapes(1);

    CreateConstraintList();

    // Start an ambient friction effect/
    hlBeginFrame();
    hlEffectd(HL_EFFECT_PROPERTY_GAIN, config::phantomFrictionGain);
    hlEffectd(HL_EFFECT_PROPERTY_MAGNITUDE, config::phantomFrictionMagnitude);
    hlStartEffect(HL_EFFECT_FRICTION, friction);
    hlEndFrame();
}

//////////////////////////////////////////////////////////////////////

MazeGraphicsHaptics::~MazeGraphicsHaptics() {
    printf("deleting MazeGraphicsHaptics (%x)\n", this);
	hlDeleteShapes(mazeListId, 1);
	hlDeleteShapes(constraintListId, 1);
}

//////////////////////////////////////////////////////////////////////

void MazeGraphicsHaptics::Update(double elapsed) {
    uint32 timeOut = 5000;

    // Make sure we have the workspace set correctly before any haptics
	DrawGraphics();
	UpdateWorkspace();
    DrawHaptics();

    // Force phantom to start position
    if (!atStart) {
        double pos[3];
        hlGetDoublev(HL_DEVICE_POSITION, pos);

        if (maze->isEntrance(pos)) {
            // Stay here for a while to let the phantom settle
            uint32 startTime = sysGetTime();
            uint32 lastTime = sysGetTime();
            while (lastTime - startTime < timeOut) {
                lastTime = sysGetTime();
            }
            atStart = true;
            hlEnable(HL_PROXY_RESOLUTION);
        }
        else {
            double start[2];
            maze->getStart(start);

            double startPos[3];
            startPos[0] = start[0];
            startPos[1] = playerHeight;
            startPos[2] = start[1];

            hlBeginFrame();

            hlDisable(HL_PROXY_RESOLUTION);
            hlProxydv(HL_PROXY_POSITION, startPos);

            hlEndFrame();
        }
    }

	HLerror error;

	while (HL_ERROR(error = hlGetError())) {
		printf("HL Error: %s\n", error.errorCode);

		if (error.errorCode == HL_DEVICE_ERROR) {
			hduPrintError(stderr, &error.errorInfo,
								"Error during haptic rendering\n");
		}
	}
}

//////////////////////////////////////////////////////////////////////

void MazeGraphicsHaptics::CreateConstraintList() {
    // Constrain player between two planes.  Need to do this with two contact planes
    // instead of one constraint plane because there seems to be a bug in OpenHL that
    // allows passing through contact objects when on a constraint...
    const double constraintHeight = 0.001;

	glDeleteLists(constraintList, 1);
    constraintList = glGenLists(1);
    glNewList(constraintList, GL_COMPILE);

    glBegin(GL_QUADS);
        // Ceiling, CCW
        glVertex3f(0, playerHeight + constraintHeight, 0);
        glVertex3f(maze->getWidth(), playerHeight + constraintHeight, 0);
        glVertex3f(maze->getWidth(), playerHeight + constraintHeight, maze->getHeight());
        glVertex3f(0, playerHeight + constraintHeight, maze->getHeight());

        // Floor, CW
        glVertex3f(0, playerHeight - constraintHeight, 0);
        glVertex3f(0, playerHeight - constraintHeight, maze->getHeight());
        glVertex3f(maze->getWidth(), playerHeight - constraintHeight, maze->getHeight());
        glVertex3f(maze->getWidth(), playerHeight - constraintHeight, 0);
    glEnd();

    glEndList();

}

//////////////////////////////////////////////////////////////////////

void MazeGraphicsHaptics::DrawHaptics() {
    // Start haptic frame
    hlBeginFrame();

    // Set material properties for the shapes
    hlMaterialf(HL_FRONT, HL_STIFFNESS, 0.9);
    hlMaterialf(HL_FRONT, HL_DAMPING, 0.1);
    hlMaterialf(HL_FRONT, HL_STATIC_FRICTION, 0.2);
    hlMaterialf(HL_FRONT, HL_DYNAMIC_FRICTION, 0.3);

    // Draw maze
    hlTouchModel(HL_CONTACT);
    hlTouchableFace(HL_FRONT);
    hlBeginShape(HL_SHAPE_FEEDBACK_BUFFER, mazeListId);
    glCallList(mazeList);
    hlEndShape();

    // Draw constraint
    hlTouchModel(HL_CONTACT);
    hlTouchableFace(HL_FRONT);
    hlBeginShape(HL_SHAPE_FEEDBACK_BUFFER, constraintListId);
    glCallList(constraintList);
    hlEndShape();

    // End the haptic frame
    hlEndFrame();
}

//////////////////////////////////////////////////////////////////////

void MazeGraphicsHaptics::UpdateWorkspace() {
    GLdouble projection[16];
    glGetDoublev(GL_PROJECTION_MATRIX, projection);

    hlMatrixMode(HL_TOUCHWORKSPACE);
    hlLoadIdentity();

    // fit haptic workspace to view volume
    hluFitWorkspace(projection);
}

//////////////////////////////////////////////////////////////////////

#endif