//////////////////////////////////////////////////////////////////////
// Includes //////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

#include "fonts.h"

#include <stdio.h>
#include <stdarg.h>
#include "config.h"

//////////////////////////////////////////////////////////////////////
// TFont /////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

TFont::TFont(const char * filename) {
  width = config::fontWidth;
  height = config::fontHeight;

  base = glGenLists(256);

  fontTex = TTextureList::Loader(filename, false);

  fastDraw = false;
  stateReady = false;

  for (int loop1=0; loop1<256; loop1++) {
    float cx = float(loop1%16)/16.0f;
    float cy = float(loop1/16)/16.0f;  

    glNewList(base+loop1,GL_COMPILE);
    glBegin(GL_QUADS);
      glTexCoord2f(cx,           1.0f - cy);            glVertex2i(0, 0);
      glTexCoord2f(cx + 0.0625f, 1.0f - cy);            glVertex2i(1, 0);
      glTexCoord2f(cx + 0.0625f, 1.0f - cy - 0.0625f);  glVertex2i(1, 1);
      glTexCoord2f(cx,           1.0f - cy - 0.0625f);  glVertex2i(0, 1);
    glEnd();
    glTranslated(config::fontPitch/config::fontWidth, 0, 0);

    glEndList();
  }
}

//////////////////////////////////////////////////////////////////////

TFont::~TFont() {
  // Delete the display lists
  glDeleteLists(base, 256);
}

//////////////////////////////////////////////////////////////////////

void TFont::drawText(GLint x, GLint y, int set, const char *fmt, ...)  {
  char text[1024];
  char *p;
  va_list ap;
  float w, h;

  if (fmt == NULL) return;

  // Parse the argument list to construct a display string
  va_start(ap, fmt);
  vsprintf(text, fmt, ap);
  va_end(ap);

  // Justify the text
  if (x < 0) {
    getTextExtents(text, &w, &h);
    
    if (x == CENTER_JUSTIFY)
      x = (screenWidth - w) * 0.5f;
    else if (x == LEFT_JUSTIFY)
      x = 0;
    else if (x == RIGHT_JUSTIFY)
      x = screenWidth - w;
  }

  // Justify the text in the y direction
  if (y == CENTER_JUSTIFY)
    y = (screenHeight - h) * 0.5f;
  else if (y == TOP_JUSTIFY)
    y = 0;
  else if (y == BOTTOM_JUSTIFY)
    y = screenHeight - height;
  
  // Save the modelview matrix and position for text
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glTranslatef((float)x, (float)y, 0.0f);
  glScalef(width, height, 1.0f);

  // Switch character sets if needed
  if (set) set = 128;
  for (p = text; *p; p++) *p = *p ^ set;

  // Draw the text using GL display lists
  if (!fastDraw) lockState();
  glListBase(base);
  glCallLists((GLsizei)j_stringLength(text), GL_UNSIGNED_BYTE, text);
  if (!fastDraw) unlockState();

  // Restore the matrix stack
  glPopMatrix();
}

//////////////////////////////////////////////////////////////////////

void TFont::drawTrueText(GLint x, GLint y, int set, const char *fmt, ...)  {
  char text[1024];
  char *p;
  va_list ap;

  if (fmt == NULL) return;

  // Parse the argument list to construct a display string
  va_start(ap, fmt);
  vsprintf(text, fmt, ap);
  va_end(ap);
  
  // Save the modelview matrix and position for text
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glTranslatef((float)x, (float)y, 0.0f);
  glScalef(width, height, 1.0f);

  // Switch character sets if needed
  if (set) set = 128;
  for (p = text; *p; p++) *p = *p ^ set;

  // Draw the text using GL display lists
  if (!fastDraw) lockState();
  glListBase(base);
  glCallLists((GLsizei)j_stringLength(text), GL_UNSIGNED_BYTE, text);
  if (!fastDraw) unlockState();

  // Restore the matrix stack
  glPopMatrix();
}

//////////////////////////////////////////////////////////////////////

void TFont::lock(void) {
  fastDraw = true;
  lockState();
}

//////////////////////////////////////////////////////////////////////

void TFont::unlock(void) {
  fastDraw = false;
  unlockState();
}

//////////////////////////////////////////////////////////////////////

void TFont::lockState(void) {
  if (!stateReady) {
    stateReady = true;
    glEnable(GL_TEXTURE_2D);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    fontTex->activate();
  }
}

//////////////////////////////////////////////////////////////////////

void TFont::unlockState(void) {
  if (stateReady) {
    stateReady = false;
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
  }
}

//////////////////////////////////////////////////////////////////////

void TFont::getTextExtents(const char * text, float * width, float * height) {
  *width = getTextWidth(text);
  *height = getTextHeight(text);
}

//////////////////////////////////////////////////////////////////////

float TFont::getTextHeight(const char * text) {
  return config::fontHeight;
}

//////////////////////////////////////////////////////////////////////

float TFont::getTextWidth(const char * text) {
  return j_stringLength(text) * config::fontPitch;
}

//////////////////////////////////////////////////////////////////////
// Ill-placed globals ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

TFont * qfont;

//////////////////////////////////////////////////////////////////////

