//////////////////////////////////////////////////////////////////////

#include "loadTGA.h"

//////////////////////////////////////////////////////////////////////

// TARGA Format:
//  size  name         description
//  -----------------------------
//   18   header       File header (see TTGAHeader)
//  [var] description  ASCII ID field (size is header.idSize)
//  [var] palette      Palette entries (present if header.paletteType <> 0, 
//                     the size (in bytes) is header.paletteLength *
//                                            ceil(header.paletteEntrySize/8))
//  [var] image data   pixel data (size depends on compression)
//  [var] developer    Developer directory (present if footer.developerOffset > 0)
//  [var] extension    Extension area (present if footer.extensionOffset > 0)
//  [26]  footer       File footer (see TTGAFooter)

// For more information, google for TARGA.PDF

//////////////////////////////////////////////////////////////////////

#define TGA_NO_PALETTE 0
#define TGA_HAS_PALETTE 1

#define TGA_IMAGE_NONE 0
#define TGA_IMAGE_INDEXED 1
#define TGA_IMAGE_RGB 2
#define TGA_IMAGE_GREY 3

#define TGA_IMAGE_RLE 8

//////////////////////////////////////////////////////////////////////

typedef struct {
  byte  idSize;             // size of ID field that follows 18 byte header (0 usually)
  byte  paletteType;        // type of colour map 0=none, 1=has palette
  byte  imageType;          // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

  int16 paletteStart;       // first colour map entry in palette
  int16 paletteLength;      // number of colours in palette
  byte  paletteEntrySize;   // number of bits per palette entry (combined) 15,16,24,32

  int16 xstart;             // image x origin
  int16 ystart;             // image y origin
  int16 width;              // image width in pixels
  int16 height;             // image height in pixels
  byte  bpp;                // image bits per pixel 8,16,24,32
  byte  descriptor;         // image descriptor bits (vh flip bits, todo)
} TTGAHeader;

// Note: this is totally untested, if TGA output is implemented, the
// TTGAExtension and TTGAFooter structs should be verified.  I just
// quickly typed them up from targa.pdf

typedef struct {
  int16 size;
  char  author[41];
  char  comments[324];
  int16 timeStamp[6];
  char  jobName[41];
  int16 jobTime[3];
  char  softwareID[41];
  byte  softwareVer[3];
  int32 keyColor;
  int16 aspectRatio[2];
  int16 gammaValue[2];
  int32 correctionOffset;
  int32 postageOffset;
  int32 scanlineOffset;
  byte  attributeType;
  // scanline table
  // postage stamp image
  // color correction table (int16[256][4])
} TTGAExtension;

typedef struct {
  int32 extensionOffset;    // extension area offset (not included if 0)
  int32 developerOffset;    // developer directory offset (not included if 0)
  byte  signature[18];      // "TRUEVISION-XFILE.\0"
} TTGAFooter;

//////////////////////////////////////////////////////////////////////

bool TTGATexture::loadTGA(TVFSStream * stream, const char *filename) {
  // Deallocate any existing texture associated with this object
  release();
 
  // Read the file into a buffer
  byte * buffer;
  int bufferSize = stream->getSize();
  if (bufferSize < sizeof(TTGAHeader)+1) return false;
  buffer = new byte[bufferSize];
  stream->Read(buffer, bufferSize);

  // Alias the header
  TTGAHeader * header = (TTGAHeader *)buffer;

  // Localize the used header entries
  int width = localizeLEPair(header->width);
  int height = localizeLEPair(header->height);
  int stride = header->bpp >> 3;
  int imageOffset = localizeLEPair(header->paletteLength)*
                    ((header->paletteEntrySize+7)>>3) +
                    header->idSize + sizeof(TTGAHeader);

  // Verify it is valid, supported TGA
  bool valid = true;
  if (header->paletteType != TGA_NO_PALETTE) valid = false;
  if (header->imageType /*&~ TGA_IMAGE_RLE*/ != TGA_IMAGE_RGB) valid = false;
  if ((header->bpp != 24) && (header->bpp != 32)) valid = false;
  if ((width < 1) || (height < 1)) valid = false;
  if (imageOffset + width*height*stride > bufferSize) valid = false;

  // Die gracefully if needed
  if (!valid) {
    j_printf("Error: Unable to load unsupported or invalid TGA image %s\n", filename);
    delete [] buffer;
    return false;
  }

  // Get us a buffer
  if (!initialize(width, height, header->bpp)) {
    j_printf("Error: Unable to obtain a texture buffer for TGA image %s\n", filename);
    delete [] buffer;
    return false;
  }

  // todo: read the image descriptor bits 4 and 5 (zero scale)
  //   b5 b4 where first pixel ends up
  //   0  0  bottom left
  //   0  1  bottom right
  //   1  0  top left
  //   1  1  top right

  // Copy through, flipping the color order from BGR to RGB
  byte * src = &(buffer[imageOffset]);
  for (int i = 0; i < width*height*stride; i += stride) {
    myData[i+0] = src[i+2];
    myData[i+1] = src[i+1];
    myData[i+2] = src[i+0];
    if (stride == 4) myData[i+3] = src[i+3];  // alpha
  }

  // Bind the data to an OpenGL texture handle
  upload();
  
  // Free our buffer and return triumphantly  
  delete [] buffer;
  return true;
}

//////////////////////////////////////////////////////////////////////
