#ifndef MENUS_HEADER_INCLUDE
#define MENUS_HEADER_INCLUDE

//////////////////////////////////////////////////////////////////////

#include "system.h"
#include "viewpoint.h"
#include <vector>
#include <list>
#include <string>
#include "speech.h"

//////////////////////////////////////////////////////////////////////

class jMenuItem;
class MenuSystem;

typedef void (*jMenuEvent)(jMenuItem *, MenuSystem *, int);

typedef std::list<jMenuItem *> jMenuList;

//////////////////////////////////////////////////////////////////////

class jMenuItem /*: public ReferenceCounted*/ {
  friend class MenuSystem;
protected:
  // The short name is used in parent menus, the long name as a header inside submenus
  std::string * myShortName;
  int myTag;

  // Children of this menu item
  jMenuList children;

  // Event handlers
  jMenuEvent myEvent;  // called when a menu item is activated / entered
public:
  std::string * myLongName;
  // Create a new jMenuItem
  jMenuItem(const char * shortName, const char * longName)
  : //ReferenceCounted(),
    myEvent(NULL),
    myTag(-1)
  {
    myShortName = new std::string(shortName);
    myLongName = new std::string(longName);
  }

  ~jMenuItem() {
     printf("deleting jMenuItem (%x)\n", this);

    // Dereference all my children
    for (jMenuList::iterator i = children.begin(); i != children.end(); i++) {
      delete (*i);//->removeReference();
    }
    delete myShortName;
    delete myLongName;
  }

  // Add a sub-menu item
  void add(jMenuItem * child, int tag = 0) {
    children.push_back(child);
//    child->addReference();
    child->myTag = tag;
  }

  // Add an event callback
  void setEvent(jMenuEvent event) { myEvent = event; }

  void activate(MenuSystem * system) {
    if (myEvent != NULL) {
      //printf("%p menu item being activated (short: %s, long: %s, tag: %i)\n", 
      //  this, this->myShortName->c_str(), this->myLongName->c_str(), myTag);
      // Not a sub-menu, so speak the name in confirmation
      startSpeechEvent(myLongName->c_str());

      // Perform the action
      myEvent(this, system, myTag);
    }
  }

  // Add a functional menu item
  void add(const char * shortName, jMenuEvent callback, int tag = 0, const char * longName = "") {
    jMenuItem * child = new jMenuItem(shortName, longName);
    child->setEvent(callback);
    add(child, tag);
  }
};

//////////////////////////////////////////////////////////////////////

class mLeftRightWidget : public jMenuItem {
private:
  float myMin;
  float myMax;
  float myDelta;
  float * myTarget;
public:
  mLeftRightWidget(const char * shortName, float * target, float minValue = 0.0f, float maxValue = 1.0f, float delta = 0.2f)
    : myTarget(target), myMin(minValue), myMax(maxValue), myDelta(delta), jMenuItem(shortName, "")
  {}

  // do something with it
};

//////////////////////////////////////////////////////////////////////

// Owns the memory of everything in the root under it (well, menu items
// are reference counted, but it derefs everything, so as long as a
// menu graph does not span more than one menusystem, killing the root
// of the graph kills all the memory too)
class MenuSystem : TRenderObject {
protected:
  // History of menu choices to make the back button work across
  // non-tree shaped menus (probably needlessly complicated, I could
  // have constrained it to a tree and just used parent pointers!)
  std::vector<jMenuItem *> menuHistory;

  // Currently active menu and menu item
  jMenuItem * activeItem;
  jMenuItem * activeMenu;
  jMenuItem * rootMenu;
  float myTimeout, myIdleTimeout;
  void changeActiveMenu(jMenuItem * newItem);
  void changeActiveItem(jMenuItem * newItem);
public:
  bool initialize(jMenuItem * root);
  virtual void release(void);
  virtual void render(TViewpoint * context, float elapsed);

  // Go back up the menu tree
  static void goBack(jMenuItem * sender, MenuSystem * system, int tag);
  void reactivateMenu(void);
};

//////////////////////////////////////////////////////////////////////

#endif