//////////////////////////////////////////////////////////////////////

#include "speech.h"
#include "sound.h"
#include "jfilesystem.h"
#include <list>

//////////////////////////////////////////////////////////////////////

using namespace std;

//////////////////////////////////////////////////////////////////////

#define MS_SPEECH_ENABLED

#ifdef MS_SPEECH_ENABLED

#include <sapi.h>

ISpVoice * voice = NULL;

#endif

//////////////////////////////////////////////////////////////////////

class SpeechEvent {
protected:
  int speechFlags;
  Channel * sound;     // fmod handle for sampled speech
  ULONG streamNumber;  // sapi handle for synth speech
public:
  string myText;
  bool started;

  bool pollForCompletion(void);
  void start(void);

  SpeechEvent(const char * text);
  ~SpeechEvent();
};

//////////////////////////////////////////////////////////////////////

typedef list<SpeechEvent *> SpeechEventList;
SpeechEventList events;

//////////////////////////////////////////////////////////////////////

SpeechEvent::SpeechEvent(const char * text)
: myText(text),
  sound(NULL),
  streamNumber(0),
  started(false)
{
  // Try to find a file of the text
  char * prefixes[2] = { "speech\\", "" };
  char * extensions[2] = { ".wav", ".mp3" };
  
  for (int j = 0; j < 2; j++) {
    for (int i = 0; i < 2; i++) {
      string probe = prefixes[j] + myText;
      if (myText.find(extensions[i]) == string::npos) probe.append(extensions[i]);
      if (vfsFileExists(probe.c_str())) {
        sound = soundLoad(probe.c_str());
        break;
      }
    }
  }

  if (sound == NULL) {
#ifdef MS_SPEECH_ENABLED
    // Look for a text file to speak instead
    speechFlags = SPF_ASYNC;
    if (strstr(text, "<") != NULL) speechFlags |= SPF_IS_XML;
    string probe = myText;
    probe.append(".txt");
    if (vfsFileExists(probe.c_str())) {
      speechFlags = speechFlags|SPF_IS_FILENAME;
      myText = probe;
    }
#endif
  }

  printf("speech event: %s\n", text);
}

//////////////////////////////////////////////////////////////////////

void SpeechEvent::start(void) {
  if (started) return;
  started = true;

  if (sound == NULL) {
    // Start talking with some TTS magic
#ifdef MS_SPEECH_ENABLED
    static WCHAR wideText[256];
    printf("speaking %s: \n", myText.c_str());
    MultiByteToWideChar(CP_ACP, 0, myText.c_str(), (int)myText.length()+1, wideText, (int)sizeof(wideText)/sizeof(wideText[0]));
    if (voice) {
      voice->Speak(wideText, speechFlags, &streamNumber);
    }
#endif
  } else {
    // Start playing the loaded sample
    sound->start();
    sound->setVolume(256);
  }
}

//////////////////////////////////////////////////////////////////////

SpeechEvent::~SpeechEvent() {
  printf("deleting speech (%x)\n", this);
  if (sound != NULL) {
    sound->stop();
//    sound->removeReference();
    delete sound;
  } else {
#ifdef MS_SPEECH_ENABLED
    voice->Speak(L"", SPF_PURGEBEFORESPEAK, NULL);
#endif
  }
  sound = NULL;
}

//////////////////////////////////////////////////////////////////////

bool SpeechEvent::pollForCompletion(void) {
 static SPVOICESTATUS status;
  if (sound != NULL) {
    //printf("not done %s\n", myText.c_str());
    return started && !(sound->isPlaying());
  } else {
#ifdef MS_SPEECH_ENABLED
    if (voice != NULL) {
      voice->GetStatus(&status, NULL);
      return started && (status.dwRunningState == SPRS_DONE);
    }   
#endif
    return started;
  }
}

//////////////////////////////////////////////////////////////////////

// Creates a new speech event using the text passed in
// text is first treated as a filename, by appending .wav and .mp3,
// and if that fails, it is treated as text for a TTS engine
SpeechToken startSpeechEvent(const char * text) {
  SpeechEvent * event = new SpeechEvent(text);
  events.push_back(event);
  return (SpeechToken)event;
}

//////////////////////////////////////////////////////////////////////

// kills the speech event represented by the token (or fails silently
// if that speech event has already been completed)
void stopSpeechEvent(SpeechToken token) {
  SpeechEvent * event = (SpeechEvent *)token;
  if (event != NULL) {
    events.remove(event);
    delete event;
  }
}

//////////////////////////////////////////////////////////////////////

// called when someone hits CTRL or similar, kills all speech
void stopSpeechEvents(void) {
  for (SpeechEventList::iterator i = events.begin(); i != events.end(); i++) {
    delete *i;
  }
  events.clear();    
}

//////////////////////////////////////////////////////////////////////

void stopOneSpeechEvent(void) {
#ifdef MS_SPEECH_ENABLED
    voice->Speak(L"", SPF_PURGEBEFORESPEAK, NULL);
#endif
}

//////////////////////////////////////////////////////////////////////

void pollSpeechEvents(void) {
  for (SpeechEventList::iterator i = events.begin(); i != events.end(); ) {
    SpeechEventList::iterator current = i;
    i++;
    if ((*current)->pollForCompletion()) {
      printf("deleting sound event %s: \n", (*current)->myText.c_str());
      delete *current;
      events.erase(current);
    }
  }

  // Start playing a new one
  if (!events.empty()) {
    (*events.begin())->start();
  }
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

#ifdef MS_SPEECH_ENABLED

//////////////////////////////////////////////////////////////////////

void speechInit(void) {
  HRESULT hr;
  voice = NULL;

  // Initialize COM
  CoInitialize(NULL);

  // Create the voice instance
  hr = CoCreateInstance(CLSID_SpVoice, NULL, CLSCTX_ALL, IID_ISpVoice, (void**)&voice);
  if (!SUCCEEDED(hr)) voice = NULL;

  //if (voice) voice->Speak(L"Speech Ready", SPF_ASYNC, NULL);
}

//////////////////////////////////////////////////////////////////////

void speechCleanup(void) {
  // Safely release the voice interface
  if (voice != NULL ) voice->Release();
  voice = NULL;

  // Shutdown COM
  CoUninitialize();
}

//////////////////////////////////////////////////////////////////////

#else

void speechInit(void) { }
void speechCleanup(void) { }

#endif

//////////////////////////////////////////////////////////////////////
