//////////////////////////////////////////////////////////////////////
// Includes //////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

#include "system.h"
#include "vectors.h"
#include "math.h"

//////////////////////////////////////////////////////////////////////
// Constructors / destructors ////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

TVector::TVector(void) {}

//////////////////////////////////////////////////////////////////////

TVector::TVector(const TVector &b) {
  a[0] = b[0];
  a[1] = b[1];
  a[2] = b[2];
}

//////////////////////////////////////////////////////////////////////

TVector::TVector(float x, float y, float z) {
  assign(x, y, z);
}

//////////////////////////////////////////////////////////////////////

TVector::~TVector(void) {}

//////////////////////////////////////////////////////////////////////

TVector::TVector(float * v) {
  a[0] = v[0];
  a[1] = v[1];
  a[2] = v[2];
}

//////////////////////////////////////////////////////////////////////
// Property accessors ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

float TVector::x(void) const {
  return a[0];
}

//////////////////////////////////////////////////////////////////////

float TVector::y(void) const {
  return a[1];
}

//////////////////////////////////////////////////////////////////////

float TVector::z(void) const {
  return a[2];
}

//////////////////////////////////////////////////////////////////////

float TVector::operator [](int n) const {
  return a[n];
}

//////////////////////////////////////////////////////////////////////
// Assignment ////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

void TVector::assign(float x, float y, float z) {
  a[0] = x;
  a[1] = y;
  a[2] = z;
}

//////////////////////////////////////////////////////////////////////

void TVector::assign(float * v) {
  a[0] = v[0];
  a[1] = v[1];
  a[2] = v[2];
}

//////////////////////////////////////////////////////////////////////

TVector& TVector::operator =(const TVector &b) {
  a[0] = b[0];
  a[1] = b[1];
  a[2] = b[2];

  return *this;
}

//////////////////////////////////////////////////////////////////////
// Scaling ///////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

TVector& TVector::operator *=(float f) {
  a[0] *= f;
  a[1] *= f;
  a[2] *= f;

  return *this;
}

//////////////////////////////////////////////////////////////////////

TVector TVector::operator *(float f) const {
  TVector result;

  result.a[0] = a[0] * f;
  result.a[1] = a[1] * f;
  result.a[2] = a[2] * f;

  return result;
}


//////////////////////////////////////////////////////////////////////

TVector& TVector::operator /=(float f) {
  a[0] /= f;
  a[1] /= f;
  a[2] /= f;

  return *this;
}

//////////////////////////////////////////////////////////////////////

TVector TVector::operator /(float f) const {
  TVector result;

  result.a[0] = a[0] / f;
  result.a[1] = a[1] / f;
  result.a[2] = a[2] / f;

  return result;
}

//////////////////////////////////////////////////////////////////////
// Addition/subtraction //////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

TVector& TVector::operator +=(const TVector &b) {
  a[0] += b[0];
  a[1] += b[1];
  a[2] += b[2];

  return *this;
}

//////////////////////////////////////////////////////////////////////

TVector TVector::operator +(const TVector &b) const {
  TVector result;

  result.a[0] = a[0] + b[0];
  result.a[1] = a[1] + b[1];
  result.a[2] = a[2] + b[2];

  return result;
}

//////////////////////////////////////////////////////////////////////

TVector& TVector::operator -=(const TVector &b) {
  a[0] -= b[0];
  a[1] -= b[1];
  a[2] -= b[2];

  return *this;
}

//////////////////////////////////////////////////////////////////////

TVector TVector::operator -(const TVector &b) const {
  TVector result;

  result.a[0] = a[0] - b[0];
  result.a[1] = a[1] - b[1];
  result.a[2] = a[2] - b[2];

  return result;
}

//////////////////////////////////////////////////////////////////////
// Multiplication ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

float TVector::dot(const TVector &b) const {
  return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}

//////////////////////////////////////////////////////////////////////

TVector TVector::cross(const TVector &b) const {
  TVector result;

  result.a[0] = a[1]*b[2] - a[2]*b[1];
  result.a[1] = a[2]*b[0] - a[0]*b[2];
  result.a[2] = a[0]*b[1] - a[1]*b[0];

  return result;
}

//////////////////////////////////////////////////////////////////////
// Other operations //////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

float TVector::length(void) const {
  return sqrtf(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]);
}

//////////////////////////////////////////////////////////////////////

float TVector::norm(void) const {
  return a[0]*a[0] + a[1]*a[1] + a[2]*a[2];
}

//////////////////////////////////////////////////////////////////////

TVector& TVector::normalize(void) {
  float f = 1.0f / length();
  a[0] *= f;
  a[1] *= f;
  a[2] *= f;

  return *this;
}

//////////////////////////////////////////////////////////////////////

TVector TVector::asUnit(void) const {
  TVector result;

  float f = 1.0f / length();
  result.a[0] = a[0] * f;
  result.a[1] = a[1] * f;
  result.a[2] = a[2] * f;

  return result;
}

//////////////////////////////////////////////////////////////////////

TVector TVector::combine(const TVector &b) const {
  TVector result;

  result.a[0] = a[0] * b[0];
  result.a[1] = a[1] * b[1];
  result.a[2] = a[2] * b[2];

  return result;
}

//////////////////////////////////////////////////////////////////////
// Misc. Routines ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

TVector polarToVector(float radius, float theta) {
  theta *= DEG_TO_RAD;
  TVector temp(cosf(theta)*radius, 0.0, sinf(theta)*radius);
  return temp;
}

//////////////////////////////////////////////////////////////////////

float vectorToHeading(TVector direction) {
  return (float)(atan2(direction.x(), direction.z()) / DEG_TO_RAD);
}

//////////////////////////////////////////////////////////////////////
// T4x4Matrix ////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// aij, i is y coordinate, j is x coordinate
#define a11 0
#define a12 4
#define a13 8
#define a14 12
#define a21 1
#define a22 5
#define a23 9
#define a24 13
#define a31 2
#define a32 6
#define a33 10
#define a34 14
#define a41 3
#define a42 7
#define a43 11
#define a44 15

//////////////////////////////////////////////////////////////////////

// Creates an exact copy of this matrix object
T4x4Matrix * T4x4Matrix::clone(void) {
  return new T4x4Matrix(this);
}

//////////////////////////////////////////////////////////////////////

// Determines whether this matrix is the same as another one
bool T4x4Matrix::equals(T4x4Matrix * b) {
  for (int i = 0; i < 16; i++)
    if (a[i] != b->a[i]) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////

// Replaces this matrix with its inverse
bool T4x4Matrix::invert(void) {
  // fixme: implement
  return false;
}

//////////////////////////////////////////////////////////////////////

// Replaces this matrix with the identity matrix
void T4x4Matrix::identity(void) {
  a[a11] = a[a22] = a[a33] = a[a44] = 1.0f;
  a[a12] = a[a21] = a[a31] = a[a41] = 0.0f;
  a[a13] = a[a23] = a[a32] = a[a42] = 0.0f;
  a[a14] = a[a24] = a[a34] = a[a43] = 0.0f;
}

//////////////////////////////////////////////////////////////////////

// Replaces this matrix with the zero matrix
void T4x4Matrix::zero(void) {
  for (int i = 0; i < 16; i++) a[i] = 0.0f;
}

//////////////////////////////////////////////////////////////////////

// Replaces this matrix with its transpose
void T4x4Matrix::transpose(void) {
#define mswap(i, j) { float t = a[i]; a[i] = a[j]; a[j] = t; }
  mswap(a12, a21)
  mswap(a13, a31)
  mswap(a14, a41)

  mswap(a23, a32)
  mswap(a24, a42)

  mswap(a34, a43)
#undef mswap
}

//////////////////////////////////////////////////////////////////////

// Updates this matrix with the product of itself and a translation matrix
void T4x4Matrix::translate(float tx, float ty, float tz) {
  // a := a * t(sx, sy, sz)
  for (int i = 0; i < 4; i++) {
    a[a14+i] = a[a11+i]*tx + a[a12+i]*ty + a[a13+i]*tz + a[a14+i];
  }
}

//////////////////////////////////////////////////////////////////////

// Updates this matrix with the product of itself and a scaling matrix
void T4x4Matrix::scale(float sx, float sy, float sz) {
  // a := a * s(sx, sy, sz)
  for (int i = 0; i < 4; i++) {
    a[i+0] *= sx;
    a[i+4] *= sy;
    a[i+8] *= sz;
  }
}

//////////////////////////////////////////////////////////////////////

// Updates this matrix with the product of itself and a rotation matrix
void T4x4Matrix::rotate(float angle, float x, float y, float z) {
  // todo: implement
}

//////////////////////////////////////////////////////////////////////

// Updates this matrix with the product of itself and another matrix
// a := a * b
void T4x4Matrix::multiply(T4x4Matrix * b) {
  // todo: implement
  T4x4Matrix c(this);
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++) {
      a[i*4+j] = c.a[i*4+j] * b->a[i+j*4];
    }
  }
}

//////////////////////////////////////////////////////////////////////

// Updates (replace or multiply) the current OpenGL matrix with this matrix
void T4x4Matrix::upload(bool replace) {
  if (replace)
    glLoadMatrixf(a);
  else
    glMultMatrixf(a);
}

//////////////////////////////////////////////////////////////////////

// Determines whether this matrix is the identity matrix
bool T4x4Matrix::isIdentity(void) {
  int i = 0;
  for (int y = 0; y < 4; y++) {
    for (int x = 0; x < 4; x++) {
      if (x == y) {
        if (a[i] != 1.0f) return false;
      } else {
        if (a[i] != 0.0f) return false;
      }
      i++;
    }
  }

  return true;
}

//////////////////////////////////////////////////////////////////////

// Determines whether this matrix is invertible
bool T4x4Matrix::isInvertible(void) {
  // fixme: implement
  return false;
}

//////////////////////////////////////////////////////////////////////

// Determines whether this matrix is orthogonal
bool T4x4Matrix::isOrthogonal(void) {
// fixme: implement
  return false;
}

//////////////////////////////////////////////////////////////////////

// Determines whether this matrix is the zero/null matrix
bool T4x4Matrix::isZero(void) {
  for (int i = 0; i < 16; i++) if (a[i] != 0.0f) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////

// Calculates the determinant of this matrix
float T4x4Matrix::calculateDeterminant(void) {
// fixme: implement
  return 0.0f;
}

//////////////////////////////////////////////////////////////////////

// Transforms an array of points using this matrix
void T4x4Matrix::transformPoints(TVector * vectors, int count) {
// fixme: implement
}

//////////////////////////////////////////////////////////////////////

// Add another matrix to this one
void T4x4Matrix::add(T4x4Matrix * b) {
  for (int i = 0; i < 16; i++) a[i] += b->a[i];
}

//////////////////////////////////////////////////////////////////////

// Subtract another matrix from this one
void T4x4Matrix::subtract(T4x4Matrix * b) {
  for (int i = 0; i < 16; i++) a[i] -= b->a[i];
}

//////////////////////////////////////////////////////////////////////

// Scale every element in this matrix by a scalar factor
void T4x4Matrix::scaleAll(float f) {
  for (int i = 0; i < 16; i++) a[i] *= f;
}

//////////////////////////////////////////////////////////////////////

// Clone this matrix
T4x4Matrix::T4x4Matrix(T4x4Matrix * b) {
  for (int i = 0; i < 16; i++) a[i] = b->a[i];
}

//////////////////////////////////////////////////////////////////////
