//////////////////////////////////////////////////////////////////////

#ifndef VECTORS_H
#define VECTORS_H

//////////////////////////////////////////////////////////////////////
// Includes //////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

#include "system.h"

// (2*pi)/360
#define DEG_TO_RAD 0.01745329252f

// 360/(2*pi)
#define RAD_TO_DEG 57.2957795130f

//////////////////////////////////////////////////////////////////////
// TVector ///////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

class TVector {
public:
  float a[3];

  // Constructors / destructors
  TVector::TVector(void);
  TVector::TVector(const TVector &b);
  TVector::TVector(float x, float y, float z);
  TVector::~TVector(void);
  TVector::TVector(float * v);

  // Access
  float TVector::x(void) const;
  float TVector::y(void) const;
  float TVector::z(void) const;
  float TVector::operator [](int n) const;

  // Assignment
  void TVector::assign(float x, float y, float z);
  void TVector::assign(float * v);
  TVector& TVector::operator =(const TVector &b);

  // Scaling
  TVector& TVector::operator *=(float f);
  TVector TVector::operator *(float f) const;
  TVector& TVector::operator /=(float f);
  TVector TVector::operator /(float f) const;

  // Addition/subtraction
  TVector& TVector::operator +=(const TVector &b);
  TVector TVector::operator +(const TVector &b) const;
  TVector& TVector::operator -=(const TVector &b);
  TVector TVector::operator -(const TVector &b) const;

  // Multiplication
  float TVector::dot(const TVector &b) const;
  TVector TVector::cross(const TVector &b) const;

  // Other operations
  float TVector::length(void) const;
  float TVector::norm(void) const;
  TVector& TVector::normalize(void);
  TVector TVector::asUnit(void) const;
  TVector TVector::combine(const TVector &b) const;
};

//////////////////////////////////////////////////////////////////////

// This class encapsulates a matrix stored in column-major order:
//   a0  a4  a8  a12
//   a1  a5  a9  a13
//   a2  a6  a10 a14
//   a3  a7  a11 a15
class T4x4Matrix {
public:
  float a[16];
  
  // Creates an exact copy of this matrix object
  T4x4Matrix * clone(void);

  // Determines whether this matrix is the same as another one
  bool equals(T4x4Matrix * b);

  // Replaces this matrix with its inverse
  bool invert(void);

  // Replaces this matrix with the identity matrix
  void identity(void);

  // Replaces this matrix with the zero matrix
  void zero(void);

  // Replaces this matrix with its transpose
  void transpose(void);

  // Updates this matrix with the product of itself and a translation matrix
  void translate(float tx, float ty, float tz);

  // Updates this matrix with the product of itself and a scaling matrix
  void scale(float sx, float sy, float sz);

  // Updates this matrix with the product of itself and a rotation matrix
  void rotate(float angle, float x, float y, float z);

  // Updates this matrix with the product of itself and another matrix
  void multiply(T4x4Matrix * b);

  // Updates (replace or multiply) the current OpenGL matrix with this matrix
  void upload(bool replace);

  // Determines whether this matrix is the identity matrix
  bool isIdentity(void);

  // Determines whether this matrix is invertible
  bool isInvertible(void);

  // Determines whether this matrix is orthogonal
  bool isOrthogonal(void);

  // Determines whether this matrix is the zero/null matrix
  bool isZero(void);

  // Calculates the determinant of this matrix
  float calculateDeterminant(void);

  // Transforms an array of points using this matrix
  void transformPoints(TVector * vectors, int count);

  // Add another matrix to this one
  void add(T4x4Matrix * b);

  // Subtract another matrix from this one
  void subtract(T4x4Matrix * b);

  // inner product

  // Scale every element in this matrix by a scalar factor
  void scaleAll(float f);

  // Clone this matrix
  T4x4Matrix(T4x4Matrix * b);
};

//////////////////////////////////////////////////////////////////////
// Misc. Routines ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

TVector polarToVector(float radius, float theta);
float vectorToHeading(TVector direction);

//////////////////////////////////////////////////////////////////////

#endif

//////////////////////////////////////////////////////////////////////
