//////////////////////////////////////////////////////////////////////
unit mazeTypes; //////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
interface ////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

uses
  Windows, SysUtils, StdCtrls, ComObj, Classes, Contnrs, Graphics,
  Dialogs, Math;

//////////////////////////////////////////////////////////////////////

var
  voice: OLEVariant;
  voiceGood: boolean;

//////////////////////////////////////////////////////////////////////

const
  MATERIAL_FLOOR = 0;
  MATERIAL_WALL = 1;
  MATERIAL_ENTRANCE = 2;
  MATERIAL_EXIT = 3;

//////////////////////////////////////////////////////////////////////

type
  TCell = class
  public
    material: integer;  // will be uid of material, but isn't right now
    zone: integer;      // uid of zone
    constructor Create;
    destructor Destroy; override;
    procedure clear;
  end;
         
  TArrayOfCells = array of TCell;

//////////////////////////////////////////////////////////////////////

  TMapResource = class
  public
    uid: integer;
    name: string;
    color: TColor;
    constructor Create; virtual;
    procedure LoadFromFile(var fp: TextFile); virtual; abstract;
    procedure SaveToFile(var fp: TextFile); virtual; abstract;
  end;

  TMapResourceList = class
  protected
    elements: TStringList;
  public
    function GetUniqueName(prefix: string): string;
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromFile(var fp: TextFile); virtual; abstract;
    procedure SaveToFile(var fp: TextFile); virtual;
    function Find(id: integer): TMapResource;
    procedure Add(resource: TMapResource);
    procedure Rename(index: integer; const newName: string; list: TListBox = nil);

    procedure Clear;

    // Removes the resource from the list but does not free it
    procedure Remove(resource: TMapResource);

    // Removes the resource from the list and deallocates it
//    procedure Delete;

    procedure UpdateDisplay(list: TListBox);
  end;

//////////////////////////////////////////////////////////////////////

  TZone = class(TMapResource)
  public
    procedure LoadFromFile(var fp: TextFile); override;
    procedure SaveToFile(var fp: TextFile); override;
  end;

  TZoneList = class(TMapResourceList)
  public
    procedure LoadFromFile(var fp: TextFile); override;
  end;

//////////////////////////////////////////////////////////////////////

  TEntity = class(TMapResource)
  public
    x, y, z: single;
    triggered: boolean;
    procedure Assign(source: TEntity); virtual;
    procedure LoadFromFile(var fp: TextFile); override;
    procedure SaveToFile(var fp: TextFile); override;
  end;

  TEntityList = class(TMapResourceList)
  public
    procedure LoadFromFile(var fp: TextFile); override;

    // Finds the closest entity to (x,y) that is still within
    // a distance radius from (x,y), or return nil
    // (z is ignored, hence column)
    function FindInColumn(x, y, radius: single): TEntity;
  end;

//////////////////////////////////////////////////////////////////////

  TMazeSound = class(TEntity)
  public
    volume, cutoff, radius: single;
    filename: string;

    constructor Create; override;
    procedure LoadFromFile(var fp: TextFile); override;
    procedure SaveToFile(var fp: TextFile); override;
    procedure Assign(source: TEntity); override;
  end;

//////////////////////////////////////////////////////////////////////

  TMaze = class
  private
    myWidth, myHeight: integer;
    myCellSize: single;
    myCells: TArrayOfCells;
    FModified: boolean;
    FOnChangeMod: TNotifyEvent;
    procedure SetModified(const Value: boolean);
  public
    scale: integer;
    zones: TZoneList;
    entities: TEntityList;
    description: string;

    constructor Create(width, height: integer; cellSize: single);
    destructor Destroy; override;

    procedure Resize(newWidth, newHeight: integer);

    function LoadFromFile(const filename: string): boolean;
    procedure SaveToFile(const filename: string);

    function ReadCell(x, y, z: integer): integer;
    procedure WriteCell(x, y, z, c: integer);

    property width: integer read myWidth;
    property height: integer read myHeight;
    property cellSize: single read myCellSize;
    property modified: boolean read FModified write SetModified;

    property OnChangeModified: TNotifyEvent read FOnChangeMod write FOnChangeMod;

    procedure Render(canvas: TCanvas);
    procedure DrawGlyph(canvas: TCanvas; x, y: integer; const glyph: string);
  end;

  TUndoableAction = class
  public
    sequenceID: integer;
    procedure Undo(maze: TMaze); virtual; abstract;
    procedure Redo(maze: TMaze); virtual; abstract;
    class function Name: string; virtual; abstract;
  end;

//////////////////////////////////////////////////////////////////////

function getNewUID: integer;

//////////////////////////////////////////////////////////////////////

const
  INVALID_UID = -1;

//////////////////////////////////////////////////////////////////////
implementation ///////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

var
  uidStream: integer;

//////////////////////////////////////////////////////////////////////

function getNewUID: integer;
begin
  Inc(uidStream);
  Result := uidStream;
end;

//////////////////////////////////////////////////////////////////////
// TMaze /////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

constructor TMaze.Create(width, height: integer; cellSize: single);
var
  x, y, i: integer;
begin
  uidStream := 0;

  // Initialization
  scale := 16;
  FOnChangeMod := nil;
  FModified := false;
  myWidth := width;
  myHeight := height;
  myCellSize := Max(cellSize, 0.01);

  // Clear the map
  SetLength(myCells, myWidth*myHeight);
  for i := myWidth * myHeight - 1 downto 0 do
    myCells[i] := TCell.Create;

  // Frame the map
  for x := 0 to myWidth - 1 do begin
    myCells[x].material := MATERIAL_WALL;
    myCells[x+(myHeight-1)*myWidth].material := MATERIAL_WALL;
  end;
  for y := 0 to myHeight - 1 do begin
    myCells[y*myWidth].material := MATERIAL_WALL;
    myCells[y*myWidth+myWidth-1].material := MATERIAL_WALL;
  end;

  // Create the lists
  zones := TZoneList.Create;
  entities := TEntityList.Create;
end;

//////////////////////////////////////////////////////////////////////

destructor TMaze.Destroy;
begin
  // Delete the lists
  entities.Free;
  zones.Free;
end;

//////////////////////////////////////////////////////////////////////

procedure TMaze.DrawGlyph(canvas: TCanvas; x, y: integer; const glyph: string);
var
  size: TSize;
begin
  // Draw a circle behind the text
  canvas.Pen.Color := clBlack;
  canvas.Ellipse(Rect(x, y, x+scale, y+scale));

  // Figure out where to put the text
  size := canvas.TextExtent(glyph);
  Inc(x, (scale - size.cx) div 2);
  Inc(y, (scale - size.cy) div 2);

  // Draw the text
  canvas.Pen.Mode := pmXor;
  canvas.Font.Color := clWhite;
  canvas.Brush.Style := bsClear;
  canvas.TextOut(x, y, glyph);

  // Restore canvas state (somewhat)
  canvas.Brush.Style := bsSolid;
  canvas.Pen.Mode := pmCopy;
end;

//////////////////////////////////////////////////////////////////////

function TMaze.LoadFromFile(const filename: string): boolean;
var
  fp: TextFile;
  x, y, width, height: integer;
  line: string;
begin
  Result := false;

  if not FileExists(filename) then Exit;

  // Open the file
  AssignFile(fp, filename);
  Reset(fp);
  try
    // Read the maze version
    ReadLn(fp, line);
    if line <> 'AMAZEV.2' then begin
      CloseFile(fp);
      Exit;
    end;

    // Read the maze width and height
    ReadLn(fp, width, height);
    Resize(width, height);

    // Read the maze contents
    for y := 0 to height - 1 do begin
      ReadLn(fp, line);
      SetLength(line, width);
      line := Uppercase(line);
      for x := 0 to width - 1 do begin
        case line[x+1] of
          ' ': myCells[y*width+x].material := MATERIAL_FLOOR;
          'X': myCells[y*width+x].material := MATERIAL_WALL;
          'S': myCells[y*width+x].material := MATERIAL_ENTRANCE;
          'E': myCells[y*width+x].material := MATERIAL_EXIT;
        end;
      end;
    end;

    // Read the cell size
    ReadLn(fp, myCellSize);
    myCellSize := Max(myCellSize, 0.01);

    // Read zones
    zones.Clear;
    zones.LoadFromFile(fp);

    // Read the zone assignments
    for y := 0 to myHeight - 1 do begin
      for x := 0 to myWidth - 1 do begin
        Read(fp, myCells[y*myWidth+x].zone);
      end;
      ReadLn(fp);
    end;

    // Read entities, items, or sounds
    entities.Clear;
    entities.LoadFromFile(fp);

    // Read the uid sequence point
    ReadLn(fp, uidStream);

    // Read the description of the maze
    description := '';
    while not Eof(fp) do begin
      ReadLn(fp, line);
      description := description + #13#10 + line;
    end;
  finally
    // Close the file
    CloseFile(fp);
  end;

  FModified := false;
  Result := true;
end;

//////////////////////////////////////////////////////////////////////

procedure TMaze.SaveToFile(const filename: string);
var
  fp: TextFile;
  x, y: integer;
  line: string;
begin
  AssignFile(fp, filename);
  Rewrite(fp);

  // Write the maze version
  WriteLn(fp, 'AMAZEV.2');

  // Write the width and height of the map
  WriteLn(fp, myWidth, ' ', myHeight);

  // Write the map cells
  for y := 0 to myHeight - 1 do begin
    SetLength(line, myWidth);
    line := Uppercase(line);
    for x := 0 to myWidth - 1 do begin
      case myCells[y*myWidth+x].material of
        MATERIAL_FLOOR: line[x+1] := ' ';
        MATERIAL_WALL: line[x+1] := 'X';
        MATERIAL_ENTRANCE: line[x+1] := 'S';
        MATERIAL_EXIT: line[x+1] := 'E';
      else
        line[x+1] := '?';
      end;
    end;
    WriteLn(fp, line);
  end;

  // Write the cell size
  WriteLn(fp, myCellSize);

  // Write the map zones
  zones.SaveToFile(fp);

  // Write the zone assignments
  for y := 0 to myHeight - 1 do begin
    for x := 0 to myWidth - 1 do begin
      Write(fp, myCells[y*myWidth+x].zone, ' ');
    end;
    WriteLn(fp);
  end;

  // Write any entities/items/sounds in the map
  entities.SaveToFile(fp);

  // Write the uid sequence point
  WriteLn(fp, uidStream);

  // Write the description of the maze
  WriteLn(fp, description);

  // Close the file and clear the modified flag
  CloseFile(fp);
  FModified := false;
end;

//////////////////////////////////////////////////////////////////////

function TMaze.ReadCell(x, y, z: integer): integer;
begin
  Result := -1;
  if (x >= 0) and (x < myWidth) and (y >= 0) and (y < myHeight) then begin
    if z = 0 then
      Result := myCells[x+y*myWidth].material
    else if z = 1 then
      Result := myCells[x+y*myWidth].zone;
  end;
end;

//////////////////////////////////////////////////////////////////////

procedure TMaze.Render(canvas: TCanvas);
var
  i, c, x, y: integer;
  f: single;
  zone: TZone;
  entity: TEntity;
begin
  // Draw the map
  for y := 0 to myHeight - 1 do begin
    for x := 0 to myWidth - 1 do begin
      c := myCells[x+y*myWidth].material;
      case c of
        MATERIAL_FLOOR, MATERIAL_ENTRANCE, MATERIAL_EXIT: begin
          // let the zone shine through
          zone := zones.Find(myCells[x+y*myWidth].zone) as TZone;
          if zone <> nil then
            canvas.Brush.Color := zone.color
          else
            canvas.Brush.Color := clWhite;
        end;
        MATERIAL_WALL: canvas.Brush.Color := clBlack;
      else
        canvas.Brush.Color := clAqua;
      end;
      canvas.FillRect(Rect(scale*x, scale*y, scale*(x+1), scale*(y+1)));

      case c of
        MATERIAL_ENTRANCE: begin
          canvas.Brush.Color := clRed;
          DrawGlyph(canvas, x*scale, y*scale, 'S');
        end;
        MATERIAL_EXIT: begin
          canvas.Brush.Color := clBlue;
          DrawGlyph(canvas, x*scale, y*scale, 'E');
        end;
      end;

    end;
  end;

  // Draw entities
  f := scale / myCellSize;
  for i := 0 to entities.elements.Count - 1 do begin
    entity := entities.elements.Objects[i] as TEntity;
    canvas.Brush.Color := entity.color;
    DrawGlyph(canvas, Trunc(entity.x*f), Trunc(entity.y*f), '@');
  end;
end;

//////////////////////////////////////////////////////////////////////

procedure TMaze.Resize(newWidth, newHeight: integer);
var
  dest: TArrayOfCells;
  x, y: integer;
begin
  SetLength(dest, newWidth*newHeight);

  for y := 0 to min(myHeight, newHeight) - 1 do begin
    // Copy existing cells over
    for x := 0 to min(myWidth, newWidth) - 1 do
      dest[y*newWidth + x] := myCells[y*myWidth+x];

    // Clear new cells on the right-hand side (if there are any)
    for x := myWidth to newWidth - 1 do
      dest[y*newWidth + x] := TCell.Create;

    // Free destroyed cells on the right-hand side (if there are any)
    for x := newWidth to myWidth - 1 do
      myCells[y*myWidth+x].Free;
  end;

  // Clear new cells on the bottom (if there are any)
  for y := myHeight to newHeight - 1 do begin
    for x := 0 to newWidth - 1 do
      dest[y*newWidth + x] := TCell.Create;
  end;

  // Free destroyed cells on the bottom (if there are any)
  for y := newHeight to myHeight - 1 do begin
    for x := 0 to myWidth - 1 do
      myCells[y*myWidth+x].Free;
  end;

  myWidth := newWidth;
  myHeight := newHeight;
  Finalize(myCells);
  myCells := dest;
end;

//////////////////////////////////////////////////////////////////////

procedure TMaze.SetModified(const Value: boolean);
begin
  FModified := Value;
  if Assigned(FOnChangeMod) then FOnChangeMod(self);
end;

//////////////////////////////////////////////////////////////////////

procedure TMaze.WriteCell(x, y, z, c: integer);
begin
  if (x >= 0) and (x < myWidth) and (y >= 0) and (y < myHeight) then begin
    if z = 0 then
      myCells[x+y*myWidth].material := c
    else if z = 1 then
      myCells[x+y*myWidth].zone := c;
    Modified := true;
  end;
end;

//////////////////////////////////////////////////////////////////////
// TCell /////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

procedure TCell.clear;
begin
//
end;

//////////////////////////////////////////////////////////////////////

constructor TCell.Create;
begin
//
end;

//////////////////////////////////////////////////////////////////////

destructor TCell.Destroy;
begin
//
end;

//////////////////////////////////////////////////////////////////////
// TEntity ///////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

procedure TEntity.Assign(source: TEntity);
begin
  uid := source.uid;
  x := source.x;
  y := source.y;
  z := source.z;
  triggered := source.triggered;
end;

//////////////////////////////////////////////////////////////////////

{$HINTS OFF}
procedure TEntity.LoadFromFile(var fp: TextFile);
var
  skip: char;
  trig: integer;
begin
  ReadLn(fp, uid, color, x, y, z, trig, skip, name);
  triggered := trig <> 0;
end;
{$HINTS ON}

//////////////////////////////////////////////////////////////////////

procedure TEntity.SaveToFile(var fp: TextFile);
var
  trig: integer;
begin
  if triggered then trig := 1 else trig := 0;
  WriteLn(fp, uid, ' ', color, ' ', x, ' ', y, ' ', z, ' ', trig, ' ', name);
end;

//////////////////////////////////////////////////////////////////////
// TMazeSound ////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

procedure TMazeSound.Assign(source: TEntity);
var
  sound: TMazeSound;
begin
  inherited;
  if (source is TMazeSound) then begin
    sound := source as TMazeSound;
    filename := sound.filename;
    volume := sound.volume;
    cutoff := sound.cutoff;
    radius := sound.radius;
  end;
end;

//////////////////////////////////////////////////////////////////////

constructor TMazeSound.Create;
begin
  x := -1.0;
  y := -1.0;
  z := 1.0;
  triggered := false;
  volume := 128.0;
//  cutoff :=
  inherited;
end;

//////////////////////////////////////////////////////////////////////

procedure TMazeSound.LoadFromFile(var fp: TextFile);
begin
  inherited;
  ReadLn(fp, volume, cutoff, radius, filename);
  filename := Trim(filename);
end;

//////////////////////////////////////////////////////////////////////

procedure TMazeSound.SaveToFile(var fp: TextFile);
begin
  inherited;
  WriteLn(fp, volume, ' ', cutoff, ' ', radius, ' ', filename)
end;

//////////////////////////////////////////////////////////////////////
// TMapResourceList //////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

constructor TMapResourceList.Create;
begin
  elements := TStringList.Create;
end;

//////////////////////////////////////////////////////////////////////

procedure TMapResourceList.Clear;
begin
  elements.Clear;
  
end;

//////////////////////////////////////////////////////////////////////

destructor TMapResourceList.Destroy;
var
  i: integer;
begin
  for i := 0 to elements.Count - 1 do begin
    elements.Objects[i].Free;
  end;
  elements.Free;
  inherited;
end;

//////////////////////////////////////////////////////////////////////

function TMapResourceList.GetUniqueName(prefix: string): string;
var
  i: integer;
  st: string;
begin
  // Find a unique free name
  i := elements.Count + 1;
  repeat
    Result := prefix + ' ' + IntToStr(i);
  until elements.IndexOf(st) = -1;
end;

//////////////////////////////////////////////////////////////////////

procedure TMapResourceList.SaveToFile(var fp: TextFile);
var
  i: integer;
begin
  WriteLn(fp, elements.Count);
  for i := 0 to elements.Count - 1 do begin
    (elements.Objects[i] as TMapResource).SaveToFile(fp);
  end;
end;

//////////////////////////////////////////////////////////////////////

procedure TMapResourceList.Remove(resource: TMapResource);
var
  index: integer;
begin
  index := elements.IndexOfObject(resource);
  if index >= 0 then elements.Delete(index);
end;

//////////////////////////////////////////////////////////////////////

procedure TMapResourceList.Rename(index: integer; const newName: string; list: TListBox);
begin
  if (index >= 0) and (index < elements.Count) then begin
    (elements.Objects[index] as TMapResource).name := newName;
    elements.Strings[index] := newName;
    if list <> nil then list.Items[index] := newName;
  end;
end;

//////////////////////////////////////////////////////////////////////
// TZoneList /////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

function TMapResourceList.Find(id: integer): TMapResource;
var
  i: integer;
begin
  // fixme: slow hacky
  for i := 0 to elements.Count - 1 do begin
    Result := elements.Objects[i] as TMapResource;
    if Result.uid = id then Exit;
  end;
  Result := nil;
end;

//////////////////////////////////////////////////////////////////////

procedure TZoneList.LoadFromFile(var fp: TextFile);
var
  count, i: integer;
  zone: TZone;
begin
  ReadLn(fp, count);
  for i := 0 to count - 1 do begin
    zone := TZone.Create;
    zone.LoadFromFile(fp);
    elements.AddObject(zone.name, zone);
  end;
end;

//////////////////////////////////////////////////////////////////////
// TZone /////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

{$HINTS OFF}
procedure TZone.LoadFromFile(var fp: TextFile);
var
  skip: char;
begin
  ReadLn(fp, uid, color, skip, name);
end;
{$HINTS ON}

//////////////////////////////////////////////////////////////////////

procedure TZone.SaveToFile(var fp: TextFile);
begin
  WriteLn(fp, uid, ' ', color, ' ', name);
end;

//////////////////////////////////////////////////////////////////////
// TEntityList ///////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

function TEntityList.FindInColumn(x, y, radius: single): TEntity;
var
  bestDistance, dist: single;
  i: integer;
  test: TEntity;
begin
  Result := nil;
  bestDistance := 1e30;
  for i := 0 to elements.Count - 1 do begin
    test := TEntity(elements.Objects[i]);
    dist := sqr(test.x-x) + sqr(test.y-y);
    if (dist < radius) and (dist < bestDistance) then begin
      bestDistance := dist;
      Result := test;
    end;
  end;
end;

//////////////////////////////////////////////////////////////////////

procedure TEntityList.LoadFromFile(var fp: TextFile);
var
  count, i: integer;
  entity: TEntity;
begin
  ReadLn(fp, count);
  for i := 0 to count - 1 do begin
    entity := TMazeSound.Create;  // fixme: hardcoded
    entity.LoadFromFile(fp);
    elements.AddObject(entity.name, entity);
  end;
end;

//////////////////////////////////////////////////////////////////////

procedure TMapResourceList.Add(resource: TMapResource);
begin
  elements.AddObject(resource.name, resource);
end;

//////////////////////////////////////////////////////////////////////

procedure TMapResourceList.UpdateDisplay(list: TListBox);
var
  i: integer;
begin
  if Assigned(list) then begin
    list.Items.BeginUpdate;
    list.Items.Clear;
    for i := 0 to elements.Count - 1 do begin
      list.Items.AddObject(elements.Strings[i], elements.Objects[i]);
    end;
    list.Items.EndUpdate;
  end;
end;

//////////////////////////////////////////////////////////////////////
// TMapResource //////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

constructor TMapResource.Create;
begin
  // fixme: not perfect, but it'll do for now
  color := (random(192)+64) + (random(192)+64) shl 8 + (random(192)+64) shl 16;
end;

//////////////////////////////////////////////////////////////////////

initialization
  voiceGood := false;
finalization
  if voiceGood then voice := Unassigned;
end.

//////////////////////////////////////////////////////////////////////
