//////////////////////////////////////////////////////////////////////
unit resourceActions; ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
interface ////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

uses
  Math, StdCtrls, Graphics, mazeTypes;

//////////////////////////////////////////////////////////////////////

type
  ResourceDeletionAction = class(TUndoableAction)
  protected
    resource: TMapResource; // resource to be deleted
    myList: TListBox;       // observer
    owner: boolean;         // do we 'own' the object, or does someone else
  public
    constructor Create(target: TMapResource; list: TListBox);
    destructor Destroy; override;
    procedure Undo(maze: TMaze); override;
    procedure Redo(maze: TMaze); override;
    class function Name: string; override;
  end;

  ResourceRenameAction = class(TUndoableAction)
  protected
    resource: TMapResource;
    myList: TListBox;
    backup, newName: string;
  public
    constructor Create(target: TMapResource; name: string; list: TListBox);
    destructor Destroy; override;
    procedure Undo(maze: TMaze); override;
    procedure Redo(maze: TMaze); override;
    class function Name: string; override;
  end;

  ResourcePickColorAction = class(TUndoableAction)
  protected
    resource: TMapResource;
    backup, color: TColor;
  public
    constructor Create(target: TMapResource; newColor: TColor);
    destructor Destroy; override;
    procedure Undo(maze: TMaze); override;
    procedure Redo(maze: TMaze); override;
    class function Name: string; override;
  end;

//////////////////////////////////////////////////////////////////////
implementation ///////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// ResourceDeletionAction ////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

constructor ResourceDeletionAction.Create(target: TMapResource; list: TListBox);
begin
  owner := false;
  resource := target;
  myList := list;
end;

//////////////////////////////////////////////////////////////////////

destructor ResourceDeletionAction.Destroy;
begin
  if owner then resource.Free;
  inherited;
end;

//////////////////////////////////////////////////////////////////////

class function ResourceDeletionAction.Name: string;
begin
  Result := 'Delete';
end;

//////////////////////////////////////////////////////////////////////

procedure ResourceDeletionAction.Redo(maze: TMaze);
var
  index: integer;
begin
  // Take the zone back from the zone list
  if resource is TZone then
    maze.zones.Remove(resource)
  else if resource is TEntity then
    maze.entities.Remove(resource)
  else
    Exit;
  owner := true;

  // Notify the list
  if Assigned(myList) then begin
    index := myList.Items.IndexOfObject(resource);
    if index >= 0 then begin
      myList.Items.Delete(index);
      myList.ItemIndex := Min(index, myList.Items.Count - 1);
    end;
  end;

  // We just modified stuff
  maze.modified := true;
end;

//////////////////////////////////////////////////////////////////////

procedure ResourceDeletionAction.Undo(maze: TMaze);
begin
  // Give the zone to the zone list
  if resource is TZone then
    maze.zones.Add(resource)
  else if resource is TEntity then
    maze.entities.Add(resource)
  else
    Exit;
  owner := false;

  // Update the zone display
  if Assigned(myList) then
    myList.ItemIndex := myList.Items.AddObject(resource.name, resource);
end;

//////////////////////////////////////////////////////////////////////
// ResourceRenameAction //////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

constructor ResourceRenameAction.Create(target: TMapResource; name: string; list: TListBox);
begin
  resource := target;
  myList := list;
  backup := resource.name;
  newName := name;
end;

//////////////////////////////////////////////////////////////////////

destructor ResourceRenameAction.Destroy;
begin
  inherited;
end;

//////////////////////////////////////////////////////////////////////

class function ResourceRenameAction.Name: string;
begin
  Result := 'Rename';
end;

//////////////////////////////////////////////////////////////////////

procedure ResourceRenameAction.Redo(maze: TMaze);
var
  index: integer;
begin
  // Rename the resource
  backup := resource.name;
  resource.name := newName;

  // Update the display
  if Assigned(myList) then begin
    index := myList.Items.IndexOfObject(resource);
    if index >= 0 then myList.Items.Strings[index] := resource.name;
  end;

  // We just modified stuff
  maze.modified := true;
end;

//////////////////////////////////////////////////////////////////////

procedure ResourceRenameAction.Undo(maze: TMaze);
var
  index: integer;
begin
  // Restore the old name
  resource.name := backup;

  // Update the display
  if Assigned(myList) then begin
    index := myList.Items.IndexOfObject(resource);
    if index >= 0 then myList.Items.Strings[index] := resource.name;
  end;
end;

//////////////////////////////////////////////////////////////////////
// ResourcePickColorAction ///////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

constructor ResourcePickColorAction.Create(target: TMapResource; newColor: TColor);
begin
  resource := target;
  color := newColor;
end;

//////////////////////////////////////////////////////////////////////

destructor ResourcePickColorAction.Destroy;
begin
  inherited;
end;

//////////////////////////////////////////////////////////////////////

class function ResourcePickColorAction.Name: string;
begin
  Result := 'Change Color';
end;

//////////////////////////////////////////////////////////////////////

procedure ResourcePickColorAction.Redo(maze: TMaze);
begin
  // Make a backup of the color and change it to the new one
  backup := resource.color;
  resource.color := color;

  // We just modified stuff
  maze.modified := true;
end;

//////////////////////////////////////////////////////////////////////

procedure ResourcePickColorAction.Undo(maze: TMaze);
begin
  resource.color := backup;
end;

//////////////////////////////////////////////////////////////////////

end.

//////////////////////////////////////////////////////////////////////
